/* Copyright (C) 2017 Mono Wireless Inc. All Rights Reserved.    *
 * Released under MW-SLA-*J,*E (MONO WIRELESS SOFTWARE LICENSE   *
 * AGREEMENT).                                                   */

/*
 * 本コードは超簡単！TWELITE標準アプリのUARTメッセージの解釈、及び
 * 各種コマンドの送信を行うサンプルです。
 */

#include <SoftwareSerial.h>
#include <sercmd.h>         // 書式形式のパーサー

String VersionNumber = "0.1.0";

// 受信したデータを保持する構造体
typedef struct
{
  byte LID;               // 送信側の論理デバイスID
  byte CMD;               // コマンド
  unsigned long SID;      // 送信側のシリアルナンバー
  byte IDE;               // パケット識別子
  byte PVE;               // プロトコルバージョン
  byte LQI;               // LQI
  byte RID;               // 受信側の論理デバイスID
  byte RNM;               // 中継回数
  float TMS;              // タイムスタンプ
  unsigned short PWR;     // 電源電圧(mV)
  byte DI;                // DIの状態 (1だったらLo)
  byte DIM;               // DIのマスク(1度でも変化したら1)
  unsigned short AD[4];   // アナログ電圧
} tsAppTwelite;

// Prototype
void vPutChar( unsigned char chr );    // 1バイトシリアル出力する関数を登録する。
bool bParseAppTwelite( sercmd ser, tsAppTwelite* sAppTwelite );   // 読み込んだバイト列をアナログ値などに変換する関数
void vShowData(tsAppTwelite sAppTwelite);                   // 変換したデータをPCに出力する関数

// Global Object
SoftwareSerial ser(2, 3);                   // RX, TX
sercmd SerialIn(false, 0, NULL );            // シリアル読み込み用のオブジェクト
sercmd SerialOut(false, 0, vPutChar );      // シリアル書き込み用のオブジェクト

void setup() {
  // シリアルの初期化
  Serial.begin(38400);

  // 起動メッセージの出力
  Serial.print("MONOWIRELESS App_Twelite Reader v");
  Serial.println(VersionNumber);
  Serial.println();

  // デフォルトの115200だと取りこぼすので38400に設定
  ser.begin(38400);
}

void loop() {
  /*
   * TWELITEからのシリアルデータを読み込み、PCへシリアル出力する
   */
  int c = ser.read();
  if(c >= 0){
    // １バイトずつ解釈をしていく
    byte u8stat = SerialIn.u8Parse(c);

    // 解釈に成功したら終了
    if (u8stat == E_SERCMD_COMPLETE) {
      // 0x81コマンドだったら表示
      tsAppTwelite sAppTwelite;
      if(bParseAppTwelite(SerialIn, &sAppTwelite)){
        vShowData(sAppTwelite);
      }

    } else if (u8stat == E_SERCMD_CHECKSUM_ERROR) {
      // 電文中のチェックサムと計算したチェックサムが異なっているので、おそらく正しく読み込めなかった
      Serial.print("Command checksum error\n");
    }
  }

  /*
   * PCからのシリアルデータを読み込み、入力された文字によってTWELITEにコマンドを送信する
   */
  int d = Serial.read();  // PCからの入力を読み込む
  // 入力があった場合、処理する
  if(d >= 0){
    unsigned char cmdtmp[] = {0x78,0x80,0x01,0xFF,0x0F,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF}; // シリアルメッセージのテンプレート

    // 入力値によってメッセージを変更する
    switch(d){
      case '0':
        // DIを消灯する
        cmdtmp[3] = 0x00;   // DIをすべてHiにする
        break;
      case '1':
        // DIを全灯する
        cmdtmp[3] = 0x0F;   // DIをすべてLoにする
        break;
      case '2':
        // PWMを消灯する
        cmdtmp[3] = 0x00;   // DIをすべてHiにする
        cmdtmp[4] = 0x00;   // DIの状態は変更しない
        cmdtmp[5] = 0x00; cmdtmp[6] = 0x00;   // PWM1をLoにする
        cmdtmp[7] = 0x00; cmdtmp[8] = 0x00;   // PWM2をLoにする
        cmdtmp[9] = 0x00; cmdtmp[10] = 0x00;  // PWM3をLoにする
        cmdtmp[11] = 0x00; cmdtmp[12] = 0x00; // PWM4をLoにする
        break;
      case '3':
        // PWMを全灯する
        cmdtmp[3] = 0x00;   // DIをすべてHiにする
        cmdtmp[4] = 0x00;   // DIの状態は変更しない
        cmdtmp[5] = 0x04; cmdtmp[6] = 0x00;   // PWM1をHiにする
        cmdtmp[7] = 0x04; cmdtmp[8] = 0x00;   // PWM2をHiにする
        cmdtmp[9] = 0x04; cmdtmp[10] = 0x00;  // PWM3をHiにする
        cmdtmp[11] = 0x04; cmdtmp[12] = 0x00; // PWM4をHiにする
        break;
      default:
        // 何もしない
        break;
    }

    // DIの指定部分が変わっている場合シリアルメッセージをTWELITEに送信する
    if( cmdtmp[3] != 0xFF ){
      // バイト列のコポー
      memcpy( SerialOut.au8data, cmdtmp, sizeof(cmdtmp) );
      // バイト列の長さを指定する
      SerialOut.u16len = sizeof(cmdtmp);
      // TWELITEにコマンドを送信する
      SerialOut.vOutput();
    }
  }
}

// 書式形式出力用に1バイト出力関数
void vPutChar( unsigned char chr )
{
  ser.write(chr); // 1バイトシリアル出力する関数
  Serial.print((char)chr);
}

// シリアルデータを解釈する 解釈出来たらTrue,出来なかったらFalseを返す
bool bParseAppTwelite( sercmd cmd, tsAppTwelite* sAppTwelite )
{
  sAppTwelite->LID = cmd.au8data[0];
  sAppTwelite->CMD = cmd.au8data[1];

  // コマンドIDが0x81かつバイト列が23バイトの時はIO変化のパケットなので解釈する
  if(sAppTwelite->CMD == 0x81 && cmd.u16len == 23){
    sAppTwelite->IDE = cmd.au8data[2];
    sAppTwelite->PVE = cmd.au8data[3];
    sAppTwelite->LQI = cmd.au8data[4];
    sAppTwelite->SID = ((unsigned long)cmd.au8data[5]<<24)|((unsigned long)cmd.au8data[6]<<16)|(cmd.au8data[7]<<8)|(cmd.au8data[8]);
    sAppTwelite->RID = cmd.au8data[9];
    sAppTwelite->TMS = (cmd.au8data[10]<<8|cmd.au8data[11])/64.0;
    sAppTwelite->RNM = cmd.au8data[12];
    sAppTwelite->PWR = cmd.au8data[13]<<8|cmd.au8data[14];
    sAppTwelite->DI = cmd.au8data[16];
    sAppTwelite->DIM = cmd.au8data[17];

    // アナログ値の計算を行う
    for(int i=0; i<4; i++){
      sAppTwelite->AD[i] = cmd.au8data[18+i]<<2;
      sAppTwelite->AD[i] += ( cmd.au8data[22]>>(2*i) )&0x03;
      sAppTwelite->AD[i] = sAppTwelite->AD[i]<<2;
    }
    return true;
  }
  return false;
}

// 表示する
void vShowData(tsAppTwelite sAppTwelite)
{
  if(sAppTwelite.CMD == 0x81){
    Serial.print("Logical ID = 0x");Serial.println(sAppTwelite.LID, HEX);
    Serial.print("Command ID = 0x");Serial.println(sAppTwelite.CMD, HEX);
    Serial.print("LQI = ");Serial.println(sAppTwelite.LQI);
    Serial.print("Serial ID = 0x");Serial.println(sAppTwelite.SID, HEX);
    Serial.print("Receive ID = 0x");Serial.println(sAppTwelite.RID, HEX);
    Serial.print("Timer = ");Serial.println(sAppTwelite.TMS);
    Serial.print("Number of Relay = ");Serial.println(sAppTwelite.RNM);
    Serial.print("Power Voltage = ");Serial.println(sAppTwelite.PWR);
    Serial.print("DI = ");
    // DIの状態ビット (普通に出力すると先頭の0が消えるため、1ビットずつ出力する)
    for(int i=3; i>=0; i-- ){
      Serial.print((sAppTwelite.DI&(1<<i))>>i, BIN);
    }
    Serial.println();

    // DIの変更状態ビット (普通に出力すると先頭の0が消えるため、1ビットずつ出力する)
    Serial.print("DI Mask = ");
    for(int i=3; i>=0; i-- ){
      Serial.print((sAppTwelite.DIM&(1<<i))>>i, BIN);
    }
    Serial.println();
    Serial.print("ADC1 = ");Serial.println(sAppTwelite.AD[0]);
    Serial.print("ADC2 = ");Serial.println(sAppTwelite.AD[1]);
    Serial.print("ADC3 = ");Serial.println(sAppTwelite.AD[2]);
    Serial.print("ADC4 = ");Serial.println(sAppTwelite.AD[3]);
    Serial.println();
  }  
}
